from typing import TYPE_CHECKING

from .ABC import SQLConstraints

if TYPE_CHECKING:
    from .Classes import EasyColumn

__all__ = ['PRIMARY', 'NOT_NULL', 'AUTO_INCREMENT', 'UNIQUE', 'Unique']

PRIMARY = SQLConstraints('PRIMARY KEY')
NOT_NULL = SQLConstraints('NOT NULL')
AUTO_INCREMENT = SQLConstraints('AUTO_INCREMENT')
UNIQUE = SQLConstraints('UNIQUE')


class Unique(SQLConstraints):
    """
    Constraint representing unique
    """

    def __init__(self, *columns: 'EasyColumn', name: str = None):
        """
        Unique constractor

        :param columns: the column or columns for this constraint
        :param name: the name for this constraint
        """
        if name is None:
            super().__init__(f'UNIQUE ({", ".join([column.name for column in columns])})')
        else:
            super().__init__(f'CONSTRAINT {name} UNIQUE ({", ".join([column.name for column in columns])})')



from abc import ABC
from typing import Callable, Any, Iterable

__all__ = ['SQLType', 'SQLConstraints', 'SQLCommand', 'SQLCommandExecutable', 'SQLExecutable',
           'CHARSET', 'make_collection', 'is_collection']


class SQLType:
    def __init__(self, name, *args, caster: Callable[[Any], Any] = None, get_caster: Callable[["SQLType"], Callable[[Any], Any]] = None, default: Any = None, parser: Callable[[Any], str] = None, modifiable: bool = False, tags: Iterable[str] = None):
        self._name = name
        self._args = args
        self._tags = tags or ()

        self._modify_args = dict(caster=caster, get_caster=get_caster, default=default, parser=parser)

        if caster is None and get_caster is not None:
            caster = get_caster(self)

        if caster is None:
            raise NotImplementedError('cast method is not implemented')

        try:
            caster(default)
        except Exception:
            raise NotImplementedError('cast method is not implemented correctly')

        self._caster = caster
        self._default = default

        self._parser = parser if parser is not None else lambda value: 'null' if value is None else str(value)
        self._modifiable = modifiable

    def __call__(self, *args):
        if self._modifiable or not args:
            return SQLType(self._name, *args, **self._modify_args, modifiable=self._modifiable)

        from EasySQL import SQLTypeException
        raise SQLTypeException('this sql type is not accepting new arguments')

    def __eq__(self, other):
        try:
            return other.name == self.name and other.args == self.args
        except Exception:
            return False

    def __hash__(self):
        return hash((self.name, self.args))

    def __repr__(self):
        return f'<SQLTYPE "{self.name}">'

    @property
    def name(self):
        name = f'{self._name}({",".join([str(arg) for arg in self._args])})' if self._args else self._name
        for tag in self._tags:
            name += f' {tag}'

        return name

    def cast(self, value):
        return self._caster(value)

    def parse(self, value):
        return self._parser(self.cast(value))

    @property
    def default(self):
        return self._default

    @property
    def args(self):
        return self._args

    @property
    def modifiable(self):
        return self._modifiable


class SQLConstraints:
    def __init__(self, value):
        self.value = value
        self.column_constraint = True


class CHARSET:
    def __init__(self, name, collation, maxlen=1, description=None):
        self._name = name
        self.__doc__ = description
        self._collation = collation
        self._maxlen = maxlen

    def __repr__(self):
        return f'<CHARSET "{self._name}">'

    def __str__(self):
        return self._name

    @property
    def name(self):
        return self._name

    @property
    def collation(self):
        return self._collation

    @property
    def maxlen(self):
        return self._maxlen


class SQLExecutable(ABC):
    def execute(self, operation, params=()):
        raise NotImplementedError


class SQLCommand(ABC):
    def get_value(self, *args, **kwargs) -> str:
        raise NotImplementedError


class SQLCommandExecutable(SQLCommand, ABC):
    def execute(self, *args, **kwargs):
        raise NotImplementedError


def make_collection(value):
    return value if is_collection(value) else [value]


def is_collection(value):
    return isinstance(value, (list, set, tuple))



from .ABC import SQLType


def _get_int_cast_(size, unsigned=False):
    minimum = (-(2 ** (size - 1))) if unsigned else 0
    maximum = (2 ** (size - 1) - 1) if unsigned else 2 ** size

    def cast(value):
        if value is None:
            return None

        value = int(value)
        if not (minimum <= value <= maximum):
            raise ValueError(f'can only accept between {minimum} and {maximum}, but got {value}')

        return value

    return cast


def _float_cast(value):
    if value is None:
        return None

    return float(value)


def _string_cast(value):
    if value is None:
        return None

    return f"{value}"


def _string_parse(value):
    if value is None:
        return 'null'

    return f"'{value}'"


INT64 = BIGINT = SQLType('BIGINT', caster=_get_int_cast_(64), default=0)
INT32 = INT = INTEGER = SQLType('INT', caster=_get_int_cast_(32), default=0)
INT24 = MEDIUMINT = SQLType('MEDIUMINT', caster=_get_int_cast_(24), default=0)
INT16 = SMALLINT = SQLType('SMALLINT', caster=_get_int_cast_(16), default=0)
INT8 = TINYINT = SQLType('TINYINT', caster=_get_int_cast_(8), default=0)

BIGINT.UNSIGNED = SQLType('BIGINT', caster=_get_int_cast_(64, True), default=0)
INTEGER.UNSIGNED = SQLType('BIGINT', caster=_get_int_cast_(32, True), default=0)
MEDIUMINT.UNSIGNED = SQLType('BIGINT', caster=_get_int_cast_(24, True), default=0)
SMALLINT.UNSIGNED = SQLType('BIGINT', caster=_get_int_cast_(16, True), default=0)

BIT = SQLType('BIT', 1, get_caster=lambda self: _get_int_cast_(self.args[0]), default=0, modifiable=True)
BOOL = SQLType('BIT', 1, caster=lambda value: None if value is None else True if value else False, default=False, parser=lambda value: '1' if value else '0')

FLOAT = SQLType('FLOAT', caster=_float_cast, default=0.0)
DOUBLE = SQLType('DOUBLE', 12, 6, caster=_float_cast, default=0.0, modifiable=True)
DEC = DECIMAL = SQLType('DECIMAL', 12, 6, caster=_float_cast, default=0.0, modifiable=True)

STRING = VARCHAR = SQLType('VARCHAR', 255, caster=_string_cast, default='', parser=_string_parse, modifiable=True)
CHAR = SQLType('CHAR', 255, caster=_string_cast, default='', parser=_string_parse, modifiable=True)

type_dict = {
    INT64: ['bigint'],
    INT32: ['int', 'integer'],
    INT24: ['mediumint'],
    INT16: ['smallint'],
    INT8: ['tinyint'],
    BIT: ['bit', 'bool', 'boolean'],
    FLOAT: ['float'],
    DOUBLE: ['double'],
    DEC: ['decimal', 'dec'],
    STRING: ['varchar'],
    CHAR: ['char']
}


def string_to_type(string: str):
    args = string[string.find('(') + 1:string.rfind(')')] if string.find('(') >= 0 else None
    string = string[:string.find('(')].lower() if string.find('(') >= 0 else string

    for key, value in type_dict.items():
        if string in value:
            if key == BIT and string != 'bit':
                return BOOL
            if not key.modifiable or args is None:
                return key
            return key(*[int(arg) for arg in args.split(',')])
    return None



from typing import Optional

from .ABC import CHARSET
from .Classes import EasyDatabase, EasyColumn
from .Types import *

__all__ = ['EasyLocalHost', 'EasyInt64Column', 'EasyInt32Column', 'EasyInt24Column', 'EasyInt16Column', 'EasyInt08Column', 'EasyCharColumn',
           'EasyBitColumn', 'EasyBoolColumn', 'EasyFloatColumn', 'EasyDoubleColumn', 'EasyDecimalColumn', 'EasyStringColumn']


class EasyLocalHost(EasyDatabase):
    _host = "127.0.0.1"
    _port = 3306
    _user = "root"
    _password = ""

    def __init__(self, database, charset: Optional[CHARSET]):
        self._database = database
        self._charset = charset if charset is not None else None

        super().__init__()


class EasyInt64Column(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, INT64)


class EasyInt32Column(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, INT32)


class EasyInt24Column(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, INT24)


class EasyInt16Column(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, INT16)


class EasyInt08Column(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, INT8)


class EasyBitColumn(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, BIT)


class EasyBoolColumn(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, BOOL)


class EasyFloatColumn(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, FLOAT)


class EasyDoubleColumn(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, DOUBLE)


class EasyDecimalColumn(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, DECIMAL)


class EasyStringColumn(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, BIGINT)


class EasyCharColumn(EasyColumn):
    def __init__(self, name: str):
        super().__init__(name, BIGINT)



from .ABC import SQLCommand


class Where(SQLCommand):
    def __init__(self, sql_string):
        self.value = sql_string

    def get_value(self) -> str:
        return f'WHERE {self.value}'

    def AND(self, other: 'Where'):
        return Where(f'({self.value} AND {other.value})')

    def OR(self, other: 'Where'):
        return Where(f'({self.value} OR {other.value})')

    def NOT(self):
        return Where(f'NOT {self.value}')

    def __and__(self, other):
        if isinstance(other, Where):
            return self.AND(other)

        raise TypeError(f"unsupported operand type(s) for &: '{type(self)}' and '{type(other)}'")

    def __or__(self, other):
        if isinstance(other, Where):
            return self.OR(other)

        raise TypeError(f"unsupported operand type(s) for |: '{type(self)}' and '{type(other)}'")

    def __invert__(self):
        return self.NOT()


class WhereIsEqual(Where):
    def __init__(self, column, value):
        super().__init__(f'{column.name} = {column.parse(value)}')


class WhereIsNotEqual(Where):
    def __init__(self, column, value):
        super().__init__(f'{column.name} <> {column.parse(value)}')


class WhereIsGreater(Where):
    def __init__(self, column, value):
        super().__init__(f'{column.name} > {column.parse(value)}')


class WhereIsGreaterEqual(Where):
    def __init__(self, column, value):
        super().__init__(f'{column.name} => {column.parse(value)}')


class WhereIsLesser(Where):
    def __init__(self, column, value):
        super().__init__(f'{column.name} < {column.parse(value)}')


class WhereIsLesserEqual(Where):
    def __init__(self, column, value):
        super().__init__(f'{column.name} =< {column.parse(value)}')


class WhereIsLike(Where):
    def __init__(self, column, value):
        super().__init__(f'{column.name} LIKE {column.parse(value)}')


class WhereIsIn(Where):
    def __init__(self, column, values):
        values = (column.parse(value) for value in values)
        super().__init__(f'{column.name} IN {values}')


class WhereIsBetween(Where):
    def __init__(self, column, a, b):
        values = (column.parse(a), column.parse(b))
        super().__init__(f'{column.name} = {values}')


__all__ = ['Where', 'WhereIsEqual', 'WhereIsNotEqual', 'WhereIsGreater', 'WhereIsLesser',
           'WhereIsGreaterEqual', 'WhereIsLesserEqual', 'WhereIsLike', 'WhereIsIn', 'WhereIsBetween']


from itertools import zip_longest
from time import sleep
from typing import Optional, Union, Any, Sequence, TypeVar, Tuple, List

import mysql.connector
from deprecated.classic import deprecated

from .ABC import SQLType, CHARSET, SQLConstraints
from .Exceptions import DatabaseConnectionException
from .Logging import logger
from .Constraints import NOT_NULL, Unique, UNIQUE, PRIMARY
from .Where import Where

__all__ = ['EasyDatabase', 'EasyTable', 'EasyColumn', 'EasyForeignColumn']


def _safe_pop(d: dict, k):
    try:
        return d.pop(k)
    except KeyError:
        return None


def _ordinal(i: int):
    if 10 < i % 100 < 20:
        return f'{i}th'
    if i % 10 == 1:
        return f'{i}st'
    if i % 10 == 2:
        return f'{i}nd'
    if i % 10 == 3:
        return f'{i}rd'
    return f'{i}th'


class EasyColumn:
    def __init__(self, name: str, sql_type: SQLType, *tags: SQLConstraints, default: Any = None, order: int = None):
        self.name = name
        self.sql_type = sql_type
        self.tags = tags
        self.default = default if default else sql_type.default if NOT_NULL in self.tags else None
        self.order = order

        # if PRIMARY in self.tags and NOT_NULL in self.tags:
        #    self.tags = (tag for tag in self.tags if tag != NOT_NULL)

        self.table = None

    def set_table(self, table):
        self.table = table

    def __hash__(self):
        return hash((self.name, self.sql_type))

    def __repr__(self):
        return f'<EasyColumn "{self.name}" of "{self.table}", type={self.sql_type.name}>'

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, EasyColumn):
            return self.name == other.name and self.sql_type == other.sql_type
        return False

    def get_sql(self):
        value = f'{self.name} {self.sql_type.name}'
        for tag in self.tags:
            value += ' ' + tag.value
        if self.default is not None:
            value += f' DEFAULT {self.sql_type.parse(self.default)}'
        return value

    def parse(self, value):
        return self.sql_type.parse(value)

    def cast(self, value):
        return self.sql_type.cast(value)


class EasyForeignColumn(EasyColumn):
    @staticmethod
    def of(column: EasyColumn, name: str = None, *tags: SQLConstraints, default: Any = None):
        if not isinstance(column.table, EasyTable):
            return TypeError('Version 3: To use this method, The table of column must be set')

        tags = (NOT_NULL, ) if NOT_NULL in tags else ()
        name = f'{column.name} of {column.table.name}' if name is None else name
        return EasyForeignColumn(name, column.table, column, *tags, default=default)

    def __init__(self, name: str, table: 'EasyTable', reference: Union[EasyColumn, str], *tags: SQLConstraints, default: Any = None):
        column = table.get_column(reference)
        if column is None:
            raise ValueError(f'Unable to find `{reference}` in the table')

        self.refer_table = table
        self.refer_column = column

        tags = (NOT_NULL,) if NOT_NULL in tags else ()
        super().__init__(name, column.sql_type, *tags, default=default)

    def __repr__(self):
        return f'<EasyForeignColumn "{self.name}" reference={self.refer_table.name}({self.refer_column.name})>'

    def get_sql(self):
        return EasyColumn.get_sql(self)


class EasyDatabase:
    _database: str = None
    _password: str = None
    _host: str = "127.0.0.1"
    _port: int = 3306
    _user: str = "root"

    _charset: CHARSET = None

    _auto_connect: bool = True
    _auto_connect_delay: int = 5
    
    def __init_subclass__(cls, **kwargs):
        for key in ('database', 'password', 'host', 'port', 'user', 'charset', 'auto_connect', 'auto_connect_delay'):
            setattr(cls, f'_{key}', _safe_pop(kwargs, key) or getattr(cls, f'_{key}'))

    def __init__(self, *, _force=False):
        if type(self) == EasyDatabase and not _force:
            raise TypeError('Version 3: Unable to instance \'EasyDatabase\' directly, Create a subclass')

        if self._database is None:
            raise ValueError('database argument is required.')
        if self._password is None:
            raise ValueError('password is not provided.')
        if self._charset is not None and not isinstance(self._charset, CHARSET):
            raise TypeError(f'charset must be type of "CHARSET" or "NONE", not "{type(self._charset)}"')

        self._connection = None
        self._safe = True

        self.set_charset(self._charset)

    def _connect(self, *, attempt=1):
        while self._auto_connect or attempt == 1:
            try:
                logger.info(f'Attempting to make a connection to database \'{self._database}\' on \'{self._host}\'({_ordinal(attempt)} attempt)')
                self._connection = mysql.connector.connect(host=self._host, port=self._port, database=self._database, user=self._user, password=self._password)
                if self._connection.is_connected():
                    logger.info(f'Connection was successful')
                    break
                else:
                    raise Exception('unknown reason...')

            except Exception as e:
                logger.warn(f'Connection failed due {e}')

                if self._auto_connect:
                    sleep(self._auto_connect_delay)
            finally:
                attempt += 1

    @property
    def safe(self):
        return self._safe

    def remove_safety(self, *, confirm: bool):
        self._safe = not confirm

    @property
    def connection(self):
        if self._connection is None or not self._connection.is_connected():
            self._connect()

        if self._connection is None or not self._connection.is_connected():
            raise DatabaseConnectionException('Database is not connected')

        return self._connection

    @property
    def cursor(self):
        self._cursor = self.connection.cursor()
        return self._cursor

    @property
    def buffered_cursor(self):
        self._cursor = self.connection.cursor(buffered=True)
        return self._cursor

    @property
    def charset(self):
        return self._charset

    @property
    def name(self):
        return self._database

    def execute(self, operation, params=(), buffered=False, auto_commit=True):
        cursor = self.buffered_cursor if buffered else self.cursor

        logger.debug(f'SQL command has been requested to be executed:\n\tCommand: "{operation}"\n\tParameters: {params}\n\tCommit: {auto_commit}\tBuffered: {buffered}')
        cursor.execute(operation, params)
        if auto_commit:
            self.commit()

        return cursor

    def commit(self):
        return self.connection.commit()

    def describe_table(self, table: 'EasyTable'):
        from EasySQL.Types import string_to_type

        result = self.execute(f'DESCRIBE {self.name}.{table.name};', buffered=True).fetchall()
        columns = []
        for column in result:
            sqltype = string_to_type(column[1])
            if sqltype is None:
                raise TypeError(f'Unable to recognize name "{column[1]}" as a SQLType')

            tags = []
            prim = []
            if column[2] == 'NO':
                tags.append(NOT_NULL)
            if column[3] == 'PRI':
                prim.append(column[0])

            columns.append(EasyColumn(column[0], sqltype, *tags, default=column[4]))

        return tuple(columns)

    def set_charset(self, charset: CHARSET):
        if charset is not None:
            try:
                try:
                    command = f'SELECT DEFAULT_COLLATION_NAME, DEFAULT_CHARACTER_SET_NAME FROM information_schema.SCHEMATA WHERE information_schema.SCHEMATA.SCHEMA_NAME = \'{self.name}\''
                    col, cha = self.execute(command, auto_commit=False).fetchall()[0]
                except Exception:
                    col, cha = (None, None)

                if charset.name != cha or charset.collation != col:
                    command = f'ALTER DATABASE {self._database} CHARACTER SET = {charset.name} COLLATE = {charset.collation};'
                    self.execute(command)

                self._charset = charset

            except Exception as e:
                logger.warn(f"Altering the charset of database failed due {e}")


T = TypeVar('T')
SOS = Union[T, Sequence[T]]
ECOS = Union[EasyColumn, str]
SOS_ECOS = SOS[ECOS]


class EasyTable:
    _database: EasyDatabase = NotImplemented
    _name: str = NotImplemented
    _columns: tuple = ()

    PRIMARY: List[EasyColumn] = None
    UNIQUES: List[Unique] = None

    def __init_subclass__(cls, **kwargs):
        for key in ('database', 'name'):
            setattr(cls, f'_{key}', _safe_pop(kwargs, key) or getattr(cls, f'_{key}'))

        cls.PRIMARY = []
        cls.UNIQUES = []

        columns: List[EasyColumn] = [value for value in cls.__dict__.values() if isinstance(value, EasyColumn)]
        for column in columns:
            if UNIQUE in column.tags:
                cls.UNIQUES.append(Unique(column))
            if PRIMARY in column.tags:
                cls.PRIMARY.append(column)

            column.tags = tuple([tag for tag in column.tags if tag != UNIQUE and tag != PRIMARY])

        cls._columns: Tuple[EasyColumn] = tuple(columns)

    def __init__(self, auto_prepare: bool = True, *, _force=False):
        if type(self) == EasyTable and not _force:
            raise TypeError('Version 3: Unable to instance \'EasyTable\' directly, Create a subclass')
        
        if not isinstance(self._database, EasyDatabase):
            raise TypeError('Version 3: Database is not implemented')

        if not isinstance(self._name, str):
            raise TypeError('Version 3: Name is not implemented')
            
        self.__prepared = False

        if auto_prepare:
            self.prepare()

    def assert_columns(self, columns: SOS_ECOS) -> Optional[Sequence[EasyColumn]]:
        if columns is None or columns == '*':
            return None
        if not isinstance(columns, Sequence):
            columns = (columns,)

        return tuple(self.get_column(column, force=True) for column in columns)

    def prepare(self, alter_columns=True):
        command = f'SHOW TABLES FROM {self._database.name} WHERE Tables_in_{self._database.name} = \'{self._name}\';'
        exists = bool(self._database.execute(command, buffered=True).fetchall())
        if not exists:
            if self._columns:
                command = ', '.join([column.get_sql() for column in self._columns])
                if len(self.PRIMARY) > 0:
                    command += f", PRIMARY KEY({', '.join(column.name for column in self.PRIMARY)})"

                for column in self._columns:
                    if isinstance(column, EasyForeignColumn):
                        command += f", FOREIGN KEY ({column.name}) REFERENCES {column.refer_table.name}({column.refer_column.name})"

                for unique in self.UNIQUES:
                    command += f", {unique.value}"

                command = f"CREATE TABLE {self._name} ({command});"
                self._database.execute(command)
            else:
                raise ValueError('No columns where specified and table does not exist')
        else:
            columns = self._database.describe_table(self)
            if self._columns is None or len(self._columns) == 0:
                self._columns = columns
            else:
                c1 = set(self._columns)
                c2 = set(columns)

                if c1 != c2:
                    lc1 = [column.__repr__() for column in c1 - c2]
                    lc2 = [column.__repr__() for column in c2 - c1]
                    lc = zip_longest(lc1, lc2, "")
                    length = len(max(['Provided: '] + lc1, key=lambda col: len(col)))

                    logger.warn(f'Columns specified do not match with existing ones:\n\tProvided:{" " * (length - 10)}\t\tExisting:\n\t' +
                                '\n\t'.join([f'{lci[0]}{" " * (length - len(str(lci[0])))}\t\t{lci[1]}' for lci in lc]))
                    raise ValueError('Existing table does not match with specified columns.')
                
        self.__prepared = True

        for column in self._columns:
            column.set_table(self)

    @property
    def columns(self):
        return self._columns

    @property
    def name(self):
        return self._name
    
    @property
    def prepared(self):
        return self.__prepared

    def count_rows(self):
        return int(self._database.execute(f"SELECT COUNT(*) FROM {self.name};", buffered=True).fetchone()[0])

    def get_column(self, target: Union[ECOS], *, force=False) -> Optional[EasyColumn]:
        if target in self._columns:
            return target
        for column in self._columns:
            if column.name == target:
                return column

        if not force:
            return None
        raise ValueError(f'"{target}" is not implemented in the table({self.name}).')

    def select(self, columns: SOS_ECOS = None, where: Where = None, limit: int = None, offset: int = None, order: SOS_ECOS = None, descending: bool = False, force_one=False):
        from .Commands import Select

        assert self.prepared, 'Unable to perform action before preparing the table'
        return Select(self._database, self, self.assert_columns(columns) if columns is not None else None, where, limit, offset, self.assert_columns(order), descending, force_one).execute()

    def insert(self, columns: SOS_ECOS, values: SOS[Any], update_on_dup: bool = False):
        from .Commands import Insert

        assert self.prepared, 'Unable to perform action before preparing the table'
        return Insert(self._database, self, self.assert_columns(columns) if columns is not None or columns == '*' else self._columns, values, update_on_dup).execute()

    def update(self, columns: SOS_ECOS, values: SOS[Any], where: Where = None):
        from .Commands import Update

        assert self.prepared, 'Unable to perform action before preparing the table'
        return Update(self._database, self, self.assert_columns(columns) if columns is not None else self._columns, values, where).execute()

    def delete(self, where: Where = None):
        from .Commands import Delete

        assert self.prepared, 'Unable to perform action before preparing the table'
        return Delete(self._database, self, where).execute()

    def set(self, columns: SOS_ECOS, values: SOS[Any], where: Where = None):
        selection = self.select(columns, where)
        if selection is not None and len(selection) > 0:
            self.update(columns, values, where)
        else:
            self.insert(self.columns if columns is None or columns == '*' else columns, values)



from .ABC import *
from .Classes import *
from .Where import *
from .Exceptions import *
from .Constraints import *

from .Logging import enable_debug, disable_debug
from .Decorators import auto_init

from . import EasyInstances
from . import Types
from . import Characters as Charsets



import logging
import sys

logger = logging.getLogger('EasySQL')
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(formatter)
logger.addHandler(handler)


def enable_debug():
    logger.setLevel(logging.DEBUG)
    handler.setLevel(logging.DEBUG)


def disable_debug():
    logger.setLevel(logging.INFO)
    handler.setLevel(logging.INFO)



class DatabaseConnectionException(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return f'<DatabaseSafetyException "{self.message}">'

    def __str__(self):
        return self.message


class DatabaseSafetyException(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return f'<DatabaseSafetyException "{self.message}">'

    def __str__(self):
        return self.message


class SQLTypeException(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return f'<SQLTypeException "{self.message}">'

    def __str__(self):
        return self.message



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PyEasySQL",
    version="3.2.0",
    author="Ashenguard",
    author_email="Ashenguard@agmdev.xyz",
    description="SQL Database management without even a SQL line",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AGM-Studio/easysql",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    include_package_data=True,
    install_requires=["mysql-connector"],
)


from typing import Type, TypeVar, Union

from .Classes import EasyTable, EasyDatabase

__all__ = ['auto_init']

T = TypeVar('T', bound=Union[EasyDatabase, EasyTable])


def auto_init(cls: Type[T]) -> T:
    assert issubclass(cls, EasyDatabase) or issubclass(cls, EasyTable)

    return cls()



from typing import Union, List, Iterable, Any, Sequence

from .ABC import SQLCommandExecutable
from .Classes import EasyDatabase, EasyTable, EasyColumn
from .Exceptions import DatabaseSafetyException
from .Where import Where


class SelectData:
    def __init__(self, table: EasyTable, data_array: Union[tuple, list], columns: Union[tuple, list]):
        if len(data_array) != len(columns):
            raise ValueError('Data does not match the columns')

        columns = table.assert_columns(columns)

        self._table = table
        self._data = dict(zip(columns, data_array))

    def __repr__(self):
        return f'<SelectData source="{self._table.name}">'

    def get(self, column):
        col = self._table.get_column(column)

        if col is None or col not in self._data.keys():
            raise ValueError(f'Unable to find `{column}` in data')

        return self._data[col]

    def __iter__(self):
        return iter([self])

    def __len__(self):
        return len(self._data.keys())

    @property
    def data(self):
        return self._data.copy()


class EmptySelectData(SelectData):
    def __init__(self, table: EasyTable):
        super().__init__(table, [], [])

    def __iter__(self):
        return iter([])

    def __len__(self):
        return 0

    def __repr__(self):
        return f'<EmptySelectData source="{self._table.name}">'


class Select(SQLCommandExecutable):
    def __init__(self, database: EasyDatabase, table: EasyTable = None, columns: Sequence[EasyColumn] = None, where: Where = None, limit: int = None, offset: int = None, order: Iterable[EasyColumn] = None, descending: bool = False, force_one: bool = False):
        self._database = database
        self._table = table
        self._columns = columns
        self._where = where
        self._limit = limit
        self._offset = offset
        self._order = order
        self._desc = descending
        self._force_one = force_one

    def get_value(self) -> str:
        sql = f"SELECT {', '.join([column.name for column in self._columns]) if self._columns else '*'} FROM {self._table.name}"
        if isinstance(self._where, Where):
            sql += f" {self._where.get_value()}"
        if self._limit is not None:
            sql += f" LIMIT {self._limit}"
        if self._offset is not None:
            sql += f" OFFSET {self._offset}"
        if self._order is not None:
            sql += f" ORDER BY {','.join([column.name for column in self._order])}"
        return sql + ";"

    def execute(self) -> Union[None, SelectData, List[SelectData]]:
        result = self._database.execute(self.get_value(), auto_commit=False).fetchall()
        columns = self._columns if self._columns else self._table.columns
        new_result = tuple(SelectData(self._table, item, columns) for item in result)

        if self._force_one:
            return None if len(new_result) == 0 else new_result[0]

        return EmptySelectData(self._table) if len(new_result) == 0 else new_result[0] if len(new_result) == 1 else new_result


class Insert(SQLCommandExecutable):
    def __init__(self, database: EasyDatabase, table: EasyTable, columns: Sequence[EasyColumn], values: Sequence[Any], on_dup_update: bool = True):
        if columns == '*' or columns is None:
            columns = table.columns

        table.assert_columns(columns)
        if len(columns) != len(values):
            raise ValueError('Values length do not match with the columns of the table')

        self._database = database
        self._values = list(zip(table.columns, values))
        self._columns = columns
        self._table = table
        self._update = on_dup_update

    def get_value(self) -> str:
        columns = ', '.join([column.name for column in self._columns])
        values = ', '.join([column.parse(value) for column, value in self._values])

        if self._update:
            extra = ', '.join([f"{column.name}={column.parse(value)}" for column, value in self._values])
            extra = f" ON DUPLICATE KEY UPDATE {extra}"
        else:
            extra = ""

        return f"INSERT INTO {self._table.name} ({columns}) VALUES ({values}){extra};"

    def execute(self):
        return self._database.execute(self.get_value(), buffered=True).lastrowid


# noinspection SqlWithoutWhere
# The asserts will not allow the missing where
class Update(SQLCommandExecutable):
    def __init__(self, database: EasyDatabase, table: EasyTable, columns: Sequence[EasyColumn], values: Sequence[Any], where: Where = None):
        self._database = database
        self._columns = columns
        self._values = values
        self._table = table
        self._where = where

        if len(self._columns) != len(self._values):
            raise ValueError('Values length do not match with the columns')

    def get_value(self) -> str:
        set_command = ', '.join([f'{column.name} = {column.parse(value)}' for column, value in zip(self._columns, self._values)])
        return f"UPDATE {self._table.name} SET {set_command}" + f' {self._where.get_value()};' if self._where else ";"

    def execute(self):
        if self._database.safe and self._where is None:
            raise DatabaseSafetyException('Update without any condition is prohibited')

        return self._database.execute(self.get_value(), buffered=True).lastrowid


# noinspection SqlWithoutWhere
# The asserts will not allow the missing where
class Delete(SQLCommandExecutable):
    def __init__(self, database: EasyDatabase, table: EasyTable = None, where: Where = None):
        self._database = database
        self._table = table
        self._where = where

    def get_value(self) -> str:
        return f"DELETE FROM {self._table.name}" + f' {self._where.get_value()};' if self._where else ";"

    def execute(self):
        if self._database.safe and self._where is None:
            raise DatabaseSafetyException('Update without any condition is prohibited')

        return self._database.execute(self.get_value(), buffered=True).lastrowid



from .ABC import CHARSET

BIG5 = CHARSET('big5', 'big5_chinese_ci', 2, 'Big5 Traditional Chinese')
DEC8 = CHARSET('dec8', 'dec8_swedish_ci', 1, 'DEC West European')
CP850 = CHARSET('cp850', 'cp850_general_ci', 1, 'DOS West European')
HP8 = CHARSET('hp8', 'hp8_english_ci', 1, 'HP West European')
KOI8R = CHARSET('koi8r', 'koi8r_general_ci', 1, 'KOI8-R Relcom Russian')
LATIN1 = CHARSET('latin1', 'latin1_swedish_ci', 1, 'cp1252 West European')
LATIN2 = CHARSET('latin2', 'latin2_general_ci', 1, 'ISO 8859-2 Central European')
SWE7 = CHARSET('swe7', 'swe7_swedish_ci', 1, '7bit Swedish')
ASCII = CHARSET('ascii', 'ascii_general_ci', 1, 'US ASCII')
UJIS = CHARSET('ujis', 'ujis_japanese_ci', 3, 'EUC-JP Japanese')
SJIS = CHARSET('sjis', 'sjis_japanese_ci', 2, 'Shift-JIS Japanese')
HEBREW = CHARSET('hebrew', 'hebrew_general_ci', 1, 'ISO 8859-8 Hebrew')
TIS620 = CHARSET('tis620', 'tis620_thai_ci', 1, 'TIS620 Thai')
EUCKR = CHARSET('euckr', 'euckr_korean_ci', 2, 'EUC-KR Korean')
KOI8U = CHARSET('koi8u', 'koi8u_general_ci', 1, 'KOI8-U Ukrainian')
GB2312 = CHARSET('gb2312', 'gb2312_chinese_ci', 2, 'GB2312 Simplified Chinese')
GREEK = CHARSET('greek', 'greek_general_ci', 1, 'ISO 8859-7 Greek')
CP1250 = CHARSET('cp1250', 'cp1250_general_ci', 1, 'Windows Central European')
GBK = CHARSET('gbk', 'gbk_chinese_ci', 2, 'GBK Simplified Chinese')
LATIN5 = CHARSET('latin5', 'latin5_turkish_ci', 1, 'ISO 8859-9 Turkish')
ARMSCII8 = CHARSET('armscii8', 'armscii8_general_ci', 1, 'ARMSCII-8 Armenian')
UTF8 = CHARSET('utf8', 'utf8_general_ci', 3, 'UTF-8 Unicode')
UCS2 = CHARSET('ucs2', 'ucs2_general_ci', 2, 'UCS-2 Unicode')
CP866 = CHARSET('cp866', 'cp866_general_ci', 1, 'DOS Russian')
KEYBCS2 = CHARSET('keybcs2', 'keybcs2_general_ci', 1, 'DOS Kamenicky Czech-Slovak')
MACCE = CHARSET('macce', 'macce_general_ci', 1, 'Mac Central European')
MACROMAN = CHARSET('macroman', 'macroman_general_ci', 1, 'Mac West European')
CP852 = CHARSET('cp852', 'cp852_general_ci', 1, 'DOS Central European')
LATIN7 = CHARSET('latin7', 'latin7_general_ci', 1, 'ISO 8859-13 Baltic')
UTF8MB4 = CHARSET('utf8mb4', 'utf8mb4_general_ci', 4, 'UTF-8 Unicode')
CP1251 = CHARSET('cp1251', 'cp1251_general_ci', 1, 'Windows Cyrillic')
UTF16 = CHARSET('utf16', 'utf16_general_ci', 4, 'UTF-16 Unicode')
UTF16LE = CHARSET('utf16le', 'utf16le_general_ci', 4, 'UTF-16LE Unicode')
CP1256 = CHARSET('cp1256', 'cp1256_general_ci', 1, 'Windows Arabic')
CP1257 = CHARSET('cp1257', 'cp1257_general_ci', 1, 'Windows Baltic')
UTF32 = CHARSET('utf32', 'utf32_general_ci', 4, 'UTF-32 Unicode')
BINARY = CHARSET('binary', 'binary', 1, 'Binary pseudo charset')
GEOSTD8 = CHARSET('geostd8', 'geostd8_general_ci', 1, 'GEOSTD8 Georgian')
CP932 = CHARSET('cp932', 'cp932_japanese_ci', 2, 'SJIS for Windows Japanese')
EUCJPMS = CHARSET('eucjpms', 'eucjpms_japanese_ci', 3, 'UJIS for Windows Japanese')